@@ -15,7 +15,7 @@
       * storage code
 
     @copyright: 2000-2004 Juergen Hermann <jh@web.de>,
-                2003-2007 MoinMoin:ThomasWaldmann,
+                2003-2013 MoinMoin:ThomasWaldmann,
                 2010 Michael Foetsch <foetsch@yahoo.com>
     @license: GNU GPL, see COPYING for details.
 """
@@ -28,6 +28,9 @@ try:
 except ImportError:
     crypt = None
 
+from MoinMoin import log
+logging = log.getLogger(__name__)
+
 from MoinMoin.support.python_compatibility import hash_new, hmac_new
 
 from MoinMoin import config, caching, wikiutil, i18n, events
@@ -147,24 +150,75 @@ def getUserIdentification(request, username=None):
     return username or (request.cfg.show_hosts and request.remote_addr) or _("<unknown>")
 
 
-def encodePassword(pwd, salt=None):
-    """ Encode a cleartext password
+def encodePassword(cfg, pwd, salt=None, scheme=None):
+    """ Encode a cleartext password using the default algorithm.
 
+    @param cfg: the wiki config
     @param pwd: the cleartext password, (unicode)
-    @param salt: the salt for the password (string)
+    @param salt: the salt for the password (string) or None to generate a
+                 random salt.
+    @param scheme: scheme to use (by default will use cfg.password_scheme)
     @rtype: string
-    @return: the password in apache htpasswd compatible SHA-encoding,
-        or None
+    @return: the password hash in apache htpasswd compatible encoding,
     """
-    pwd = pwd.encode('utf-8')
+    if scheme is None:
+        scheme = cfg.password_scheme
+        configured_scheme = True
+    else:
+        configured_scheme = False
+    if scheme == '{PASSLIB}':
+        return '{PASSLIB}' + cfg.cache.pwd_context.encrypt(pwd, salt=salt)
+    elif scheme == '{SSHA}':
+        pwd = pwd.encode('utf-8')
+        if salt is None:
+            salt = random_string(20)
+        assert isinstance(salt, str)
+        hash = hash_new('sha1', pwd)
+        hash.update(salt)
+        return '{SSHA}' + base64.encodestring(hash.digest() + salt).rstrip()
+    else:
+        # should never happen as we check the value of cfg.password_scheme
+        raise NotImplementedError
 
-    if salt is None:
-        salt = random_string(20)
-    assert isinstance(salt, str)
-    hash = hash_new('sha1', pwd)
-    hash.update(salt)
 
-    return '{SSHA}' + base64.encodestring(hash.digest() + salt).rstrip()
+class Fault(Exception):
+    """something went wrong"""
+
+class NoSuchUser(Fault):
+    """raised if no such user exists"""
+
+class UserHasNoEMail(Fault):
+    """raised if user has no e-mail address in his profile"""
+
+class MailFailed(Fault):
+    """raised if e-mail sending failed"""
+
+
+def set_password(request, newpass, u=None, uid=None, uname=None,
+                 notify=False, skip_invalid=False, subject=None,
+                 text_intro=None, text_msg=None, text_data=None):
+    if uid:
+        u = User(request, uid)
+    elif uname:
+        u = User(request, auth_username=uname)
+    if u and u.exists():
+        if skip_invalid and u.enc_password == '':
+            return
+        if not newpass:
+            # set a invalid password hash
+            u.enc_password = ''
+        else:
+            u.enc_password = encodePassword(request.cfg, newpass)
+        u.save()
+        if not u.email:
+            raise UserHasNoEMail('User profile does not have an E-Mail address (name: %r id: %r)!' % (u.name, u.id))
+        if notify and not u.disabled:
+            mailok, msg = u.mailAccountData(subject=subject,
+                                            text_intro=text_intro, text_msg=text_msg, text_data=text_data)
+            if not mailok:
+                raise MailFailed(msg)
+    else:
+        raise NoSuchUser('User does not exist (name: %r id: %r)!' % (u.name, u.id))
 
 
 def normalizeName(name):
@@ -317,7 +371,7 @@ class User:
         self.recoverpass_key = ""
 
         if password:
-            self.enc_password = encodePassword(password)
+            self.enc_password = encodePassword(self._cfg, password)
 
         #self.edit_cols = 80
         self.tz_offset = int(float(self._cfg.tz_offset) * 3600)
@@ -355,7 +409,7 @@ class User:
         if not self.id:
             self.id = self.make_id()
             if password is not None:
-                self.enc_password = encodePassword(password)
+                self.enc_password = encodePassword(self._cfg, password)
 
         # "may" so we can say "if user.may.read(pagename):"
         if self._cfg.SecurityPolicy:
@@ -404,6 +458,10 @@ class User:
         """
         return os.path.exists(self.__filename())
 
+    def remove(self):
+        """ Remove user profile from disk """
+        os.remove(self.__filename())
+
     def load_from_id(self, password=None):
         """ Load user account data from disk.
 
@@ -514,44 +572,75 @@ class User:
         if not password:
             return False, False
 
-        # Check and upgrade passwords from earlier MoinMoin versions and
-        # passwords imported from other wiki systems.
-        for method in ['{SHA}', '{APR1}', '{MD5}', '{DES}']:
-            if epwd.startswith(method):
-                d = epwd[len(method):]
-                if method == '{SHA}':
-                    enc = base64.encodestring(
-                        hash_new('sha1', password.encode('utf-8')).digest()).rstrip()
-                elif method == '{APR1}':
-                    # d is of the form "$apr1$<salt>$<hash>"
-                    salt = d.split('$')[2]
-                    enc = md5crypt.apache_md5_crypt(password.encode('utf-8'),
-                                                    salt.encode('ascii'))
-                elif method == '{MD5}':
-                    # d is of the form "$1$<salt>$<hash>"
-                    salt = d.split('$')[2]
-                    enc = md5crypt.unix_md5_crypt(password.encode('utf-8'),
-                                                  salt.encode('ascii'))
-                elif method == '{DES}':
-                    if crypt is None:
-                        return False, False
-                    # d is 2 characters salt + 11 characters hash
-                    salt = d[:2]
-                    enc = crypt.crypt(password.encode('utf-8'), salt.encode('ascii'))
+        password_correct = recompute_hash = False
+        wanted_scheme = self._cfg.password_scheme
 
-                if safe_str_equal(epwd, method + enc):
-                    data['enc_password'] = encodePassword(password) # upgrade to SSHA
-                    return True, True
-                return False, False
+        # Check password and upgrade weak hashes to strong default algorithm:
+        for scheme in config.password_schemes_supported:
+            if epwd.startswith(scheme):
+                is_passlib = False
+                d = epwd[len(scheme):]
 
-        if epwd[:6] == '{SSHA}':
-            data = base64.decodestring(epwd[6:])
-            salt = data[20:]
-            hash = hash_new('sha1', password.encode('utf-8'))
-            hash.update(salt)
-            return safe_str_equal(hash.digest(), data[:20]), False
+                if scheme == '{PASSLIB}':
+                    # a password hash to be checked by passlib library code
+                    if not self._cfg.passlib_support:
+                        logging.error('in user profile %r, password hash with {PASSLIB} scheme encountered, but passlib_support is False' % (self.id, ))
+                    else:
+                        pwd_context = self._cfg.cache.pwd_context
+                        try:
+                            password_correct = pwd_context.verify(password, d)
+                        except ValueError, err:
+                            # can happen for unknown scheme
+                            logging.error('in user profile %r, verifying the passlib pw hash crashed [%s]' % (self.id, str(err)))
+                        if password_correct:
+                            # check if we need to recompute the hash. this is needed if either the
+                            # passlib hash scheme / hash params changed or if we shall change to a
+                            # builtin hash scheme (not recommended):
+                            recompute_hash = pwd_context.hash_needs_update(d) or wanted_scheme != '{PASSLIB}'
 
-        # No encoded password match, this must be wrong password
+                else:
+                    # a password hash to be checked by legacy, builtin code
+                    if scheme == '{SSHA}':
+                        d = base64.decodestring(d)
+                        salt = d[20:]
+                        hash = hash_new('sha1', password.encode('utf-8'))
+                        hash.update(salt)
+                        enc = base64.encodestring(hash.digest() + salt).rstrip()
+
+                    elif scheme == '{SHA}':
+                        enc = base64.encodestring(
+                            hash_new('sha1', password.encode('utf-8')).digest()).rstrip()
+
+                    elif scheme == '{APR1}':
+                        # d is of the form "$apr1$<salt>$<hash>"
+                        salt = d.split('$')[2]
+                        enc = md5crypt.apache_md5_crypt(password.encode('utf-8'),
+                                                        salt.encode('ascii'))
+                    elif scheme == '{MD5}':
+                        # d is of the form "$1$<salt>$<hash>"
+                        salt = d.split('$')[2]
+                        enc = md5crypt.unix_md5_crypt(password.encode('utf-8'),
+                                                      salt.encode('ascii'))
+                    elif scheme == '{DES}':
+                        if crypt is None:
+                            return False, False
+                        # d is 2 characters salt + 11 characters hash
+                        salt = d[:2]
+                        enc = crypt.crypt(password.encode('utf-8'), salt.encode('ascii'))
+
+                    else:
+                        logging.error('in user profile %r, password hash with unknown scheme encountered: %r' % (self.id, scheme))
+                        raise NotImplementedError
+
+                    if safe_str_equal(epwd, scheme + enc):
+                        password_correct = True
+                        recompute_hash = scheme != wanted_scheme
+
+                if recompute_hash:
+                    data['enc_password'] = encodePassword(self._cfg, password)
+                return password_correct, recompute_hash
+
+        # unsupported algorithm
         return False, False
 
     def persistent_items(self):
@@ -1029,44 +1118,53 @@ class User:
         if not safe_str_equal(h, parts[1]):
             return False
         self.recoverpass_key = ""
-        self.enc_password = encodePassword(newpass)
+        self.enc_password = encodePassword(self._cfg, newpass)
         self.save()
         return True
 
-    def mailAccountData(self, cleartext_passwd=None):
+    def mailAccountData(self, cleartext_passwd=None,
+                        subject=None,
+                        text_intro=None, text_msg=None, text_data=None):
         """ Mail a user who forgot his password a message enabling
             him to login again.
         """
         from MoinMoin.mail import sendmail
-        from MoinMoin.wikiutil import getLocalizedPage
         _ = self._request.getText
 
+        if not self.email:
+            return False, "user has no E-Mail address in his profile."
+
         tok = self.generate_recovery_token()
 
-        text = '\n' + _("""\
+        if subject is None:
+            subject = _('[%(sitename)s] Your wiki account data')
+        subject = subject % dict(sitename=self._cfg.sitename or "Wiki")
+        if text_intro is None:
+            text_intro = ''
+        if text_msg is None:
+            text_msg = _("""\
+Somebody has requested to email you a password recovery token.
+
+If you lost your password, please go to the password reset URL below or
+go to the password recovery page again and enter your username and the
+recovery token.
+""")
+        if text_data is None:
+            text_data = _("""\
 Login Name: %s
 
 Password recovery token: %s
 
 Password reset URL: %s?action=recoverpass&name=%s&token=%s
-""") % (
+""")
+        # note: text_intro is for custom stuff, we do not have i18n for it anyway
+        text = text_intro + '\n' + _(text_msg) + '\n' + _(text_data) % (
                         self.name,
                         tok,
                         self._request.url, # use full url, including current page
                         url_quote_plus(self.name),
                         tok, )
 
-        text = _("""\
-Somebody has requested to email you a password recovery token.
-
-If you lost your password, please go to the password reset URL below or
-go to the password recovery page again and enter your username and the
-recovery token.
-""") + text
-
-
-        subject = _('[%(sitename)s] Your wiki account data',
-                ) % {'sitename': self._cfg.sitename or "Wiki"}
         mailok, msg = sendmail.sendmail(self._request, [self.email], subject,
                                     text, mail_from=self._cfg.mail_from)
         return mailok, msg
